#  color.py
#  Copyright (c) TotallyNotSeth 2025
#  All rights reserved.

# Package imports
from pygame import Color

# Constants
WHITE = Color(255, 255, 255)
BLACK = Color(0, 0, 0)
RED = Color(255, 0, 0)
GREEN = Color(0, 255, 0)
BLUE = Color(0, 0, 255)
YELLOW = Color(255, 255, 0)
CYAN = Color(0, 255, 255)
MAGENTA = Color(255, 0, 255)



#  nuitka_build.py
#  Copyright (c) TotallyNotSeth 2025
#  All rights reserved.

# Compilation mode, standalone everywhere, except on macOS there app bundle
# nuitka-project: --mode=app
#
# Set EXE Filename
# nuitka-project: --output-filename=ttwhy
# nuitka-project: --output-dir=dist
#
# Add Package Data Files
# nuitka-project: --include-package-data=ttwhy:MxPlus_IBM_VGA_8x16.ttf
#
# Debugging options, controlled via environment variable at compile time.
# nuitka-project-if: {OS} == "Windows" and os.getenv("DEBUG_COMPILATION", "no") == "yes":
#     nuitka-project: --windows-console-mode=hide
# nuitka-project-else:
#     nuitka-project: --windows-console-mode=disable

if "__compiled__" in globals():  # If we are running in a nuitka executable
    from ttwhy.main import main
    from loguru import logger

    # noinspection PyBroadException
    try:
        main()  # Run the ttwhy demo
    except Exception:
        logger.exception("An unhandled exception occurred")
        raise
else:
    import os
    os.system("build.bat")  # Build and run the nuitka executable



#  enforce_iverilog.py
#  Copyright (c) TotallyNotSeth 2025
#  All rights reserved.

# Builtin imports
import os
import shutil
from pathlib import Path

# Package imports
from loguru import logger


class InstallError(Exception):
    """Raised when an installation fails."""
    pass


def ensure_iverilog() -> None:
    """Ensures that iverilog is installed on the system, and installs it if it is not."""
    if shutil.which("iverilog"):
        logger.debug("Icarus Verilog is already installed")
        return

    logger.info("Icarus Verilog is not installed, prompting user for installation")
    os.system(str(Path(__file__).parent.parent / "iverilog-v12-20220611-x64_setup.exe"))

    if not shutil.which("iverilog"):
        raise InstallError("Icarus Verilog installation failed")

    logger.info("Icarus Verilog installation successful")


# Function test
if __name__ == "__main__":
    while True:
        try:
            ensure_iverilog()
            break
        except InstallError:
            continue



#  character.py
#  Copyright (c) TotallyNotSeth 2025
#  All rights reserved.

# Package imports
import pygame

# Local imports
from ttwhy.color import Color, WHITE, BLACK
from ttwhy.util.helper_types import Coordinates

# Package exports
__all__ = ["Character", "BLINK"]

# Constants
BLINK = 0x1
BLINK_FRAMES = 60


class Character:
    """A class representing a terminal character with a foreground and background color."""
    def __init__(self, coords: Coordinates, char: str = " ", fg: Color = WHITE, bg: Color = BLACK,
                 font: pygame.font.Font = None) -> None:
        self.coords: Coordinates = coords
        self._char: str = char
        self.fg: Color = fg
        self.bg: Color = bg
        self.style: int = 0x0
        if not font:
            font = pygame.font.SysFont("Courier New", 32)
        self.font: pygame.font.Font = font

    @property
    def char(self) -> str:
        return self._char

    @char.setter
    def char(self, value: str = None) -> None:
        if not value or len(value) == 0:
            value = " "
        elif len(value) > 1:
            raise ValueError("`char` must be a single character")
        self._char = value

    def render(self, surface: pygame.Surface, x: int, y: int, frame_count: int = 0) -> None:
        """Render the character to a surface at a given position."""
        if frame_count < BLINK_FRAMES and BLINK & self.style:
            text = self.font.render(self.char, True, self.bg, self.bg)
        else:
            text = self.font.render(self.char, True, self.fg, self.bg)
        surface.blit(text, (x, y))

    @property
    def width(self) -> int:
        return self.font.size(" ")[0]  # Shouldn't be a problem since all characters are monospaced

    @property
    def height(self) -> int:
        return self.font.size(" ")[1]



#  nuitka_build.py
#  Copyright (c) TotallyNotSeth 2025
#  All rights reserved.

# Builtin imports
import sys
import os
from pathlib import Path

# Package imports
import pygame
from pygame.locals import QUIT
from loguru import logger

# Local imports
from ttwhy.util import ensure_iverilog
from ttwhy.character import BLINK
from ttwhy.terminal import Terminal
from ttwhy.color import *


# TERM_WIDTH = 100
# TERM_HEIGHT = 32
FPS = 165
BLINK_EVENT = pygame.event.custom_type()


def main():
    logger.add(Path(os.getenv('LOCALAPPDATA')) / "TotallyNotSeth" / "HDLGame" / "hdlgame_{time}.log",
               level="DEBUG", retention="7 days", backtrace=True, diagnose=True)

    logger.info("HDLGame")
    logger.info("Copyright (c) TotallyNotSeth 2025\n")

    # logger.debug("Ensuring Icarus Verilog is installed")
    # try:
    #     ensure_iverilog.ensure_iverilog()
    # except ensure_iverilog.InstallError:
    #     logger.exception("Failed to install Icarus Verilog")
    #     sys.exit(1)

    logger.debug("Pygame v{version} Init", version=pygame.ver)
    pygame.init()
    pygame.font.init()
    pygame.key.set_repeat(500, 50)
    clock = pygame.time.Clock()

    logger.debug("Pygame Display Init")
    screen = pygame.display.set_mode()
    pygame.mouse.set_visible(False)

    logger.debug("Terminal Init")
    font = pygame.font.Font(Path(__file__).parent / "MxPlus_IBM_VGA_8x16.ttf", 32)
    term_width = screen.get_width() // font.size(" ")[0]
    term_height = screen.get_height() // font.size(" ")[1]
    terminal = Terminal(screen, term_width, term_height, font)
    terminal.print("ChipperOS")
    # terminal.grid[1][2].style = BLINK
    terminal.print("Type \"help\" for a list of commands")
    terminal.print("")
    terminal.print("> █")

    logger.debug("Blink Event ID: {id}", id=BLINK_EVENT)

    last_mouse_pos: tuple[int, int] = (0, 0)
    current_mouse_pos: tuple[int, int]
    show_fps: bool = False
    frame_count: int = 0

    # Game Loop
    while True:
        screen.fill((0, 0, 0))

        for event in pygame.event.get():
            if event.type == QUIT:
                logger.info("Exiting HDLGame (OS request)...")
                pygame.quit()
                sys.exit()
            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    logger.info("Exiting HDLGame (ESC key pressed)...")
                    pygame.quit()
                    sys.exit()
                elif event.key == pygame.K_F2:
                    show_fps = not show_fps
                else:
                    terminal.process_input_key(event)

        # Update
        terminal.clear()
        # terminal.draw_str(" Hello, World! ", (1, 1), WHITE, RED)
        # terminal.draw_box((25, 5), 30, 10, CYAN, BLACK, DoubleBoxDrawingCharacters)
        terminal.draw_text_buffer()
        if show_fps:
            counter = f" FPS: {clock.get_fps():.2f} "
            terminal.draw_str(counter, (term_width - len(counter), 0), WHITE, BLUE)
            # terminal.print(counter)

        current_mouse_pos = terminal.surface_pos_to_char_pos(pygame.mouse.get_pos())
        terminal.set_color_at_char(last_mouse_pos, BLACK, BLACK)
        terminal.set_color_at_char(current_mouse_pos, BLACK, WHITE)
        last_mouse_pos = current_mouse_pos

        # Draw
        terminal.render(frame_count)
        frame_count += 1
        if frame_count >= FPS:
            frame_count = 0

        pygame.display.flip()
        clock.tick(FPS)


if __name__ == "__main__":
    main()



#  helper_types.py
#  Copyright (c) TotallyNotSeth 2025
#  All rights reserved.

from collections import namedtuple

# Coordinates
Coordinates = namedtuple("Coordinates", "x y")
CoordinatesType = Coordinates | tuple[int, int]


# Box drawing characters
class BoxDrawingCharacters:
    HORIZONTAL = "─"
    VERTICAL = "│"
    TOP_LEFT = "┌"
    TOP_RIGHT = "┐"
    BOTTOM_LEFT = "└"
    BOTTOM_RIGHT = "┘"
    HORIZONTAL_T = "┬"
    HORIZONTAL_B = "┴"
    VERTICAL_L = "├"
    VERTICAL_R = "┤"
    CROSS = "┼"


class StandardBoxDrawingCharacters(BoxDrawingCharacters):
    pass


class DoubleBoxDrawingCharacters(BoxDrawingCharacters):
    HORIZONTAL = "═"
    VERTICAL = "║"
    TOP_LEFT = "╔"
    TOP_RIGHT = "╗"
    BOTTOM_LEFT = "╚"
    BOTTOM_RIGHT = "╝"
    HORIZONTAL_T = "╦"
    HORIZONTAL_B = "╩"
    VERTICAL_L = "╠"
    VERTICAL_R = "╣"
    CROSS = "╬"



#  terminal.py
#  Copyright (c) TotallyNotSeth 2025
#  All rights reserved.

# Builtin imports
import typing

# Package imports
import pygame

# Local imports
from ttwhy.character import Character
from ttwhy.color import WHITE, BLACK
from ttwhy.util.helper_types import *

# Package exports
__all__ = ["Terminal"]


class Terminal:
    """A class representing a terminal with a grid of characters."""
    def __init__(self, surface: pygame.Surface, width: int = 80, height: int = 24, font: pygame.font.Font = None):
        self.surface: pygame.Surface = surface
        self.width: int = width
        self.height: int = height
        if not font:
            font = pygame.font.Font("vendored/MxPlus_IBM_VGA_8x16.ttf", 32)
        self.font: pygame.font.Font = font

        self.grid: list[list[Character]] = [
            [Character(Coordinates(x, y), font=self.font)
             for x in range(width)]
            for y in range(height)
        ]

        self.changed_chars: set[Character] = set()
        self.text_lines: list[str] = []

    def render(self, frame_count: int = 0) -> None:
        """Render the terminal to a surface."""
        for char in self.changed_chars:
            char.render(self.surface, *self.char_pos_to_surface_pos(char.coords), frame_count)

    def surface_pos_to_char_pos(self, surface_coords: CoordinatesType) -> Coordinates:
        """Convert a surface position to a character position."""
        if isinstance(surface_coords, tuple):
            surface_coords = Coordinates(*surface_coords)
        x_offset = self.surface.get_width() // 2 - self.width * self.grid[0][0].width // 2
        y_offset = self.surface.get_height() // 2 - self.height * self.grid[0][0].height // 2
        char_x = (surface_coords.x - x_offset) // self.grid[0][0].width
        char_y = (surface_coords.y - y_offset) // self.grid[0][0].height
        if char_x < 0:
            char_x = 0
        elif char_x >= self.width:
            char_x = self.width - 1
        if char_y < 0:
            char_y = 0
        elif char_y >= self.height:
            char_y = self.height - 1
        return Coordinates(char_x, char_y)

    def char_pos_to_surface_pos(self, char_coords: CoordinatesType) -> Coordinates:
        """Convert a character position to a surface position."""
        if isinstance(char_coords, tuple):
            char_coords = Coordinates(*char_coords)
        x_offset = self.surface.get_width() // 2 - self.width * self.grid[0][0].width // 2
        y_offset = self.surface.get_height() // 2 - self.height * self.grid[0][0].height // 2
        surface_x = char_coords.x * self.grid[0][0].width + x_offset
        surface_y = char_coords.y * self.grid[0][0].height + y_offset
        return Coordinates(surface_x, surface_y)

    def set_color_at_char(self, coords: CoordinatesType, fg: pygame.Color = WHITE, bg: pygame.Color = BLACK) -> None:
        """Set the background color of a character at a given position."""
        if isinstance(coords, tuple):
            coords = Coordinates(*coords)
        char = self.grid[coords.y][coords.x]
        char.fg = fg
        char.bg = bg
        self.changed_chars.add(char)

    def draw_char(self, char: str, coords: CoordinatesType, fg: pygame.Color = WHITE, bg: pygame.Color = BLACK) -> None:
        """Set a character at a given position."""
        if isinstance(coords, tuple):
            coords = Coordinates(*coords)
        char_obj = self.grid[coords.y][coords.x]
        char_obj.char = char
        self.set_color_at_char(coords, fg, bg)
        self.changed_chars.add(char_obj)

    def draw_str(self, text: str, coords: CoordinatesType, fg: pygame.Color = WHITE, bg: pygame.Color = BLACK) -> None:
        """Print text to the terminal at a given position with a given color."""
        if isinstance(coords, tuple):
            coords = Coordinates(*coords)
        for i, char in enumerate(text):
            self.draw_char(char, (coords.x + i, coords.y), fg, bg)

    def draw_box(self, coords: CoordinatesType, width: int, height: int,
                 fg: pygame.Color = WHITE, bg: pygame.Color = BLACK,
                 style: typing.Type[BoxDrawingCharacters] = StandardBoxDrawingCharacters) -> None:
        """Draw a box at a given position with a given color."""
        if isinstance(coords, tuple):
            coords = Coordinates(*coords)
        for i in range(width):
            self.draw_char(style.HORIZONTAL, (coords.x + i, coords.y), fg, bg)
            self.draw_char(style.HORIZONTAL, (coords.x + i, coords.y + height - 1), fg, bg)
        for i in range(height):
            self.draw_char(style.VERTICAL, (coords.x, coords.y + i), fg, bg)
            self.draw_char(style.VERTICAL, (coords.x + width - 1, coords.y + i), fg, bg)
        self.draw_char(style.TOP_LEFT, (coords.x, coords.y), fg, bg)
        self.draw_char(style.TOP_RIGHT, (coords.x + width - 1, coords.y), fg, bg)
        self.draw_char(style.BOTTOM_LEFT, (coords.x, coords.y + height - 1), fg, bg)
        self.draw_char(style.BOTTOM_RIGHT, (coords.x + width - 1, coords.y + height - 1), fg, bg)

    def clear(self) -> None:
        """Clear the terminal."""
        for char in self.changed_chars:
            char.char = " "
            char.bg = BLACK
            char.fg = WHITE
        self.changed_chars.clear()

    def print(self, text: str) -> None:
        """Print text to the terminal."""
        self.text_lines.append(text)
        if len(self.text_lines) > self.height:
            self.text_lines.pop(0)

    def draw_text_buffer(self) -> None:
        """Draw text to the terminal."""
        for i, line in enumerate(self.text_lines):
            self.draw_str(line, (0, i))

    def process_input_key(self, event: pygame.event.Event) -> None:
        """Process a key press."""
        line = self.text_lines[-1]
        if event.key == pygame.K_BACKSPACE:
            if len(line) > 3:
                self.text_lines[-1] = line[:-2] + "█"
        elif event.key == pygame.K_RETURN:
            self.text_lines[-1] = line[:-1]
            self.print("> █")
        else:
            self.text_lines[-1] = line[:-1] + event.unicode + "█"
